import asyncio
import json
import logging
from typing import Any, Callable

import aiohttp

from Shimarin.client.config import config
from Shimarin.client.networking import Response, send_get_request, send_post_request

logger = logging.getLogger("Shimarin")


class Event:
    def __init__(
        self,
        event_type: str,
        identifier: str,
        payload: str | bytes | None,
        session: aiohttp.ClientSession,
        server_endpoint: str = config.SERVER_ENDPOINT,
        custom_headers: dict | None = None,
    ) -> None:
        self.event_type = event_type
        self.payload = payload
        self.identifier = identifier
        self.__session = session
        self.server_endpoint = server_endpoint
        self.custom_headers = custom_headers if custom_headers else {}

    @staticmethod
    def new(
        event_data: dict[str, str],
        session: aiohttp.ClientSession,
        server_endpoint: str = config.SERVER_ENDPOINT,
        custom_headers: dict | None = None,
    ) -> "Event":
        if custom_headers is None:
            custom_headers = {}
        return Event(
            event_data["event_type"],
            event_data["identifier"],
            event_data["payload"],
            session,
            server_endpoint,
            custom_headers,
        )

    async def reply(self, payload: Any, metadata: dict[str, str] | None = None):
        if metadata is None:
            metadata = {}
        self.custom_headers["x-identifier"] = self.identifier
        self.custom_headers["X-Metadata"] = json.dumps(metadata)
        await send_post_request(
            self.__session,
            f"{self.server_endpoint}/callback",
            data=payload,
            headers=self.custom_headers,
        )

    def __str__(self) -> str:
        return f"Event: {self.event_type}, Identifier: {self.identifier}, Payload: {self.payload}"

    def __repr__(self) -> str:
        return self.__str__()


class EventHandler:
    def __init__(self, event_type: str, callback: Callable[[Any, Any], Any]) -> None:
        self.event_type = event_type
        self.callback = callback

    async def trigger(self, *args, **kwargs) -> None:
        await self.callback(*args, **kwargs)

    def __str__(self) -> str:
        return f"Event: {self.event_type}, Callback: {self.callback}"

    def __repr__(self) -> str:
        return self.__str__()


class EventsHandlers:
    def __init__(self):
        self.handlers: list[EventHandler] = []

    def register(self, handler: EventHandler) -> None:
        self.handlers.append(handler)

    def new(self, event_name: str):
        def wrapper(func: Callable):
            event_handler = EventHandler(event_name, func)
            self.register(event_handler)
            return func

        return wrapper

    async def handle(self, event: Event) -> None:
        for handle in self.handlers:
            if handle.event_type == event.event_type:
                asyncio.gather(handle.trigger(event))

    def __str__(self) -> str:
        return f"Handlers: {self.handlers}"

    def __repr__(self) -> str:
        return self.__str__()


class EventPolling:
    def __init__(self, events_handlers: EventsHandlers) -> None:
        self.session = aiohttp.ClientSession()
        self.events_handlers = events_handlers
        self.is_polling = False
        self.running_tasks: set[asyncio.Task] = set()

    async def __aenter__(self):
        await self.session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, tb):
        if self.session:
            await self.session.__aexit__(exc_type, exc_val, tb)

    async def start(
        self,
        polling_interval: int | float = 1,
        fetch: int = 10,
        custom_headers: dict | None = None,
        server_endpoint: str = config.SERVER_ENDPOINT,
        retries: int = 5,
    ):
        if custom_headers is None:
            custom_headers = {}
        self.is_polling = True
        while self.is_polling:
            await self.retry_loop(retries, server_endpoint, fetch, custom_headers)
            await asyncio.sleep(polling_interval)

    async def retry_loop(self, retries: int, endpoint: str, fetch: int, headers: dict):
        fetch_endpoint = f"{endpoint}/events?fetch={fetch}"
        for _ in range(retries):
            try:
                events: Response = await send_get_request(
                    self.session,
                    fetch_endpoint,
                    headers=headers,
                )

                if events.status == 200:
                    event_json = await events.json()
                    for event in event_json:
                        if event_json:
                            event = Event.new(event, self.session, endpoint, headers)
                            await self.__task_manager(event)
                    break
                raise aiohttp.ServerConnectionError
            except (
                aiohttp.ClientError,
                aiohttp.ServerTimeoutError,
                aiohttp.ServerDisconnectedError,
                aiohttp.ServerConnectionError,
                aiohttp.ClientConnectionError,
            ):
                logger.warning(f"Error connecting to {endpoint}! Retrying...")
                await asyncio.sleep(1)

    async def __task_manager(self, event: Event):
        task = asyncio.create_task(self.events_handlers.handle(event))
        self.running_tasks.add(task)
        task.add_done_callback(lambda t: self.running_tasks.remove(t))

    async def stop(self):
        if self.is_polling:
            self.is_polling = False
            for task in self.running_tasks:
                task.cancel()
            for task in self.running_tasks.copy():
                self.running_tasks.remove(task)



import asyncio

from Shimarin.client import events

handlers = events.EventsHandlers()


@handlers.new("update")
async def update(event: events.Event):
    print(event)
    print("replying to event")
    return await event.reply("Yes", metadata={"name": "Hello World"})


async def main():
    async with events.EventPolling(handlers) as poller:
        await poller.start(0.1, server_endpoint="http://localhost:2222")


if __name__ == "__main__":
    asyncio.run(main())



import inspect
import uuid
from datetime import datetime
from typing import IO, Callable, Literal

from Shimarin.server.exceptions import (
    CallbackIsLambdaError,
    UnknownStatusError,
)

type CallbackArguments = bytes | IO[bytes]
type CallbackMetadata = dict[str, str] | None


class Event[T]:
    def __init__(
        self,
        event_type: str,
        payload: str = "",
        callback: Callable[[CallbackArguments, CallbackMetadata], T] | None = None,
    ):
        if inspect.isfunction(callback) and callback.__name__ == "<lambda>":
            raise CallbackIsLambdaError
        self.event_type = event_type
        self.payload = payload
        self.callback = callback
        self.identifier = str(uuid.uuid1())
        self.answered = True if callback is None else False
        self.__answer: T | None = None
        self.__creation_date = datetime.now()
        self.__status: Literal["delivered", "done", "failed", "waiting"] = "waiting"

    @staticmethod
    def new(
        event_type: str, payload: str = "", callback: Callable[[CallbackArguments, CallbackMetadata], T] | None = None
    ) -> "Event":
        return Event(event_type, payload, callback)

    @property
    def age(self):
        return (datetime.now() - self.__creation_date).total_seconds()

    @property
    def answer(self) -> T | None:
        return self.__answer

    @answer.setter
    def answer(self, data):
        self.__answer = data
        self.answered = True
        self.status = "done"

    @property
    def status(self):
        return self.__status

    @status.setter
    def status(self, status_text: Literal["delivered", "done", "failed", "waiting"]):
        if status_text not in ["delivered", "done", "failed", "waiting"]:
            raise UnknownStatusError
        self.__status = status_text

    def json(self) -> dict:
        return {
            "event_type": self.event_type,
            "payload": self.payload,
            "identifier": self.identifier,
            "status": self.__status,
        }

    def __repr__(self):
        return self.json().__str__()

    async def trigger(
        self, payload: CallbackArguments, metadata: CallbackMetadata = None
    ) -> T | None:
        self.answered = True
        if self.callback is None:
            return
        if inspect.iscoroutinefunction(self.callback):
            self.answer = await self.callback(payload, metadata)
        else:
            self.answer = self.callback(payload, metadata)
        return self.answer



from abc import ABC, abstractmethod
from typing import Literal

from Shimarin.server.event import Event


class PersistenceMiddleware(ABC):

    @abstractmethod
    def register(self, ev: Event) -> None:
        raise NotImplementedError

    @abstractmethod
    def fetch(self, last=False) -> Event:
        raise NotImplementedError

    @abstractmethod
    def get(self, identifier: str) -> Event:
        raise NotImplementedError

    @abstractmethod
    def update_event_status(
        self,
        ev: Event,
        status: Literal["delivered", "done", "failed", "waiting"],
    ):
        raise NotImplementedError

    @abstractmethod
    def prune_finished(self, remove_failed=False) -> None:
        raise NotImplementedError

    @abstractmethod
    def remove(self, event_id: str) -> None:
        raise NotImplementedError



import json
import os
from dataclasses import dataclass
from typing import Callable

from flask import Blueprint, Flask, request

from Shimarin.server import events

CONTEXT_PATH = os.getenv("CONTEXT_PATH", "")


@dataclass
class LoginResponse:
    ok: bool
    status: int
    message: str

    def as_response(self):
        return {"ok": self.ok, "message": self.message}, self.status


INVALID_CREDENTIAL_RESPONSE = LoginResponse(False, 401, "Invalid credentials!")
DISABLED_RESPONSE = LoginResponse(True, 200, "Authentication disabled")
VALID_CREDENTIAL_RESPONSE = LoginResponse(True, 200, "Logged in!")


def login():
    env_user = os.getenv("SHIMARIN_USERNAME", "")
    env_pw = os.getenv("SHIMARIN_PASSWORD", "")

    if env_user == "" and env_pw == "":
        return DISABLED_RESPONSE

    username = request.headers.get("username")
    password = request.headers.get("password")

    if username == env_user and env_pw == password:
        return VALID_CREDENTIAL_RESPONSE
    else:
        return INVALID_CREDENTIAL_RESPONSE


class ShimaApp(Blueprint):
    def __init__(
        self,
        emitter: events.EventEmitter,
        login_callback: Callable[[], LoginResponse] = login,
        use_stream_response=False
    ):
        super().__init__("ShimaServer", __name__)
        self.login_callback = login_callback
        self.emitter = emitter
        self.use_stream_response = use_stream_response
        self.add_url_rule(
            CONTEXT_PATH + "/events", None, self.fetch, methods=["GET"]
        )
        self.add_url_rule(
            CONTEXT_PATH + "/callback", None, self.callback, methods=["POST"]
        )

    async def fetch(self):
        r = self.login_callback()
        if r.ok is False:
            return r.as_response()
        fetch = request.args.get("fetch")
        events_to_send = 1
        if fetch:
            events_to_send = int(fetch)
        events = []
        for _ in range(events_to_send):
            last_ev = await self.emitter.fetch_event()
            if last_ev:
                events.append(last_ev.json())
        return events

    async def callback(self):
        r = self.login_callback()
        if r.ok is False:
            return r.as_response()
        identifier = request.headers.get("x-identifier")
        metadata = json.loads(request.headers.get("X-Metadata", ""))
        payload = request.stream if self.use_stream_response else request.get_data(cache=False)
        if identifier and payload:
            await self.emitter.handle(identifier, payload, metadata)
        return {"ok": True}


if __name__ == "__main__":
    app = Flask("server")
    fa = ShimaApp(events.EventEmitter())
    app.register_blueprint(fa)
    app.run(debug=True, host="0.0.0.0", port=2222)



import json
from typing import Literal

from flask import Flask, request

from Shimarin.plugins.flask_api import ShimaApp
from Shimarin.plugins.middleware.sqlite_middleware import SQLitePersistenceMiddleware
from Shimarin.server.events import (
    CallbackArguments,
    CallbackMetadata,
    Event,
    EventEmitter,
)
from Shimarin.server.exceptions import EventAnswerTimeoutError

app = Flask("server")
emitter = EventEmitter(persistence_middleware=SQLitePersistenceMiddleware("test.db"))


def callback(params: CallbackArguments, metadata: CallbackMetadata) -> str | None:
    print(metadata)
    if metadata:
        print("Name: " + metadata.get("name", ""))
    if isinstance(params, bytes):
        return json.dumps(params.decode())


async def handle_test(params: dict = {}) -> str | Literal["fail"]:
    event = Event("update", json.dumps(params), callback)
    await emitter.send(event)
    print("waiting for answer")
    try:
        return await emitter.get_answer(
            event.identifier, timeout=60
        )  # 1 minute timeout
    except EventAnswerTimeoutError:
        return "fail"


@app.route("/test", methods=["GET"])
async def test_route():
    args = request.get_json(force=True, silent=True)
    if args is None:
        args = {}
    return await handle_test(args), 200


if __name__ == "__main__":
    emitter_app = ShimaApp(emitter)
    app.register_blueprint(emitter_app)
    app.run(debug=True, host="0.0.0.0", port=2222)



import pickle
import sqlite3
from asyncio import Lock
from pathlib import Path
from typing import Literal

from Shimarin.plugins.middleware.persistence import PersistenceMiddleware
from Shimarin.server.event import Event

lock = Lock()


class SQLitePersistenceMiddleware(PersistenceMiddleware):
    def __init__(self, db: str):
        self.database = Path(db)
        _ = self.database.is_file() and self.database.parent.mkdir(
            exist_ok=True, parents=True
        )
        self.setup()

    def setup(self):
        conn = sqlite3.connect(self.database)
        conn.execute(
            "CREATE TABLE IF NOT EXISTS events ("
            "id INTEGER PRIMARY KEY AUTOINCREMENT , "
            "identifier TEXT NOT NULL,"
            "status TEXT NOT NULL,"
            "data BLOB NOT NULL"
            ")"
        )
        conn.commit()
        conn.close()

    def register(self, ev: Event):
        conn = sqlite3.connect(self.database)
        query = "INSERT INTO events (identifier, status, data) VALUES (?, ?, ?)"
        conn.execute(query, [ev.identifier, ev.status, pickle.dumps(ev)])
        conn.commit()
        conn.close()

    def fetch(self, last=False) -> Event | None:
        conn = sqlite3.connect(self.database)
        query = (
            "SELECT * FROM events WHERE status = ? ORDER BY id "
            + ("DESC" if last else "")
            + " LIMIT 1"
        )
        cursor = conn.cursor()
        cursor.execute(query, ["waiting"])
        data = cursor.fetchone()
        cursor.close()
        event: Event | None = None
        if data:
            event = pickle.loads(data[3])
            if event:
                event.status = "delivered"
                self.update_event_status(event, "delivered")
        conn.close()
        return event

    def get(self, identifier: str) -> Event | None:
        query = "SELECT * FROM events WHERE identifier = ? LIMIT 1"
        conn = sqlite3.connect(self.database)
        cursor = conn.cursor()
        cursor.execute(query, [identifier])
        data = cursor.fetchone()
        cursor.close()
        conn.close()
        if data:
            return pickle.loads(data[3])

    def update_event_status(
        self,
        ev: Event,
        status: Literal["delivered", "done", "failed", "waiting"],
    ):
        query = "UPDATE events SET status = ?, data = ? WHERE identifier = ?"
        conn = sqlite3.connect(self.database)
        conn.execute(query, [status, pickle.dumps(ev), ev.identifier])
        conn.commit()
        conn.close()

    def prune_finished(self, remove_failed=False):
        query = "DELETE FROM events WHERE status = 'done'" + (
            " OR status = 'failed'" if remove_failed else ""
        )
        conn = sqlite3.connect(self.database)
        conn.execute(query)
        conn.commit()
        conn.close()

    def remove(self, event_id: int):
        conn = sqlite3.connect(self.database)
        query = "DELETE FROM events WHERE id = ?"
        conn.execute(query, [event_id])
        conn.commit()
        conn.close()



import json
import typing

import aiohttp


class Response:
    def __init__(self, status_code, content):
        self.status = status_code
        self.content = content

    async def json(self):
        return json.loads(self.content.decode("utf-8"))


async def send_get_request(session: aiohttp.ClientSession, *args, **kwargs) -> Response:
    """
    Function to send a rest request and return the Response
    :session A client session object from aiohttp
    :param *args and *kwargs to be used on the request
    :return a Response object with the content of the endpoint
    """
    response: typing.Union[Response, None] = None
    async with session.request("GET", *args, **kwargs) as resp:
        response = Response(resp.status, await resp.read())
    return response


async def send_post_request(
    session: aiohttp.ClientSession, *args, **kwargs
) -> Response:
    """
    Function to send a rest request and return the Response
    :session A client session object from aiohttp
    :param *args and *kwargs to be used on the request
    :return a Response object with the content of the endpoint
    """
    response: typing.Union[Response, None] = None
    async with session.request("POST", *args, **kwargs) as resp:
        response = Response(resp.status, await resp.read())
    return response



import asyncio
from datetime import datetime
from typing import Any, cast

from Shimarin.plugins.middleware.persistence import PersistenceMiddleware
from Shimarin.server.event import CallbackArguments, CallbackMetadata, Event
from Shimarin.server.exceptions import EventAnswerTimeoutError


class EventEmitter:
    def __init__(
        self,
        max_age_seconds: float = 0,
        persistence_middleware: PersistenceMiddleware | None = None,
    ):
        self.events: list[Event[Any]] = []
        self.max_age_seconds = max_age_seconds
        self.persistence_middleware = persistence_middleware
        self.cleanup_task: asyncio.Task = asyncio.create_task(self.clean_old_items())

    async def get_answer(self, event_id: str, timeout=60):
        start = datetime.now()
        ev: Event[Any] | None = None
        while True:
            await asyncio.sleep(0.1)
            if (datetime.now() - start).total_seconds() >= timeout:
                if ev:
                    ev.status = "failed"
                    if self.persistence_middleware:
                        self.persistence_middleware.update_event_status(ev, "failed")
                raise EventAnswerTimeoutError
            if self.persistence_middleware is not None:
                event = self.persistence_middleware.get(event_id)
                if event:
                    ev = cast(Event[Any], event)
            else:
                for event in self.events:
                    if event.identifier == event_id:
                        ev = cast(Event[Any], event)
            if ev != None and ev.answered:
                return cast(Any, ev.answer)

    async def clean_old_items(self):
        while True:
            for event in [x for x in self.events if x.status in ["done", "failed"]]:
                is_failed_event = event.status == "failed" or (
                    (event.age >= self.max_age_seconds)
                    if self.max_age_seconds
                    else False
                )
                if is_failed_event:
                    self.events.remove(event)
            if self.persistence_middleware:
                self.persistence_middleware.prune_finished()
            await asyncio.sleep(1)

    async def fetch_event(self, last: bool = True) -> Event | None:
        item: Event | None = None
        try:
            if self.persistence_middleware is not None:
                item = self.persistence_middleware.fetch(last)
            else:
                item = [x for x in self.events if x.status == "waiting"].pop(
                    0 if not last else -1
                )
            if item is not None:
                item.status = "delivered"
            return item
        except IndexError:
            return item

    async def send(self, event: Event) -> None:
        if self.persistence_middleware is not None:
            return self.persistence_middleware.register(event)
        self.events.append(event)

    async def handle(
        self,
        unique_identifier: str,
        payload: CallbackArguments,
        metadata: CallbackMetadata = None,
    ):
        if self.persistence_middleware is not None:
            ev = self.persistence_middleware.get(unique_identifier)
            if ev:
                response = await ev.trigger(payload, metadata)
                self.persistence_middleware.update_event_status(ev, "done")
                return response
        else:
            for event in self.events:
                if event.identifier == unique_identifier:
                    return await event.trigger(payload, metadata)



class EventAnswerTimeoutError(TimeoutError):
    def __init__(
        self, message: str = "Event timed out while waiting for answer"
    ) -> None:
        super().__init__(message)


class CallbackIsLambdaError(AssertionError):
    def __init__(
        self,
        message: str = "Callback cannot be lambda function, lambda function cannot be pickled: \n"
        + "https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled",
    ) -> None:
        super().__init__(message)


class UnknownStatusError(AssertionError):
    def __init__(
        self,
        message: str = "Invalid event status, allowed values are 'delivered', 'done', 'failed', 'waiting'",
    ) -> None:
        super().__init__(message)
